#include <stdio.h>

#include "Trie.h"
#include "Btree.h"
#include "Array.h"

extern char BtreeONLY, HtableONLY;

int BuckNum(int level, int maxlevel) {
  int i, size=1;
  for (i=maxlevel-1; i> level; i--) 
    size *=2;
  return size;
}

Boolean Trie::Insert(int keyLen, keyType * keyList, dataType d) {
  TrieNode** current=&root;
  TrieNode* tmp;
  elementType *elemptr;
  int i;
  
  if (BtreeONLY) {
    if (root==NULL) {
      *current=root= new Btree(btreeary);
    }

    for (i=0; i<keyLen; i++) {
      elemptr=(*current)->Insert(keyList[i]);
      if (i < keyLen -1 && elemptr->next==NULL) {
	elemptr->next=new Btree(btreeary);
      }
      current=&(elemptr->next);
    }
  }
  else if (HtableONLY) {
    if (root==NULL) {
      *current=root= new HashTable(BuckNum(0, MAXLEVEL), bucketsize);
    }

    for (i=0; i<keyLen; i++) {
      elemptr=(*current)->Insert(keyList[i]);
      if (i < keyLen -1 && elemptr->next==NULL) {
	elemptr->next=new HashTable(BuckNum(i+1, MAXLEVEL), bucketsize);
      }
      current=&(elemptr->next);
    }
  }
  else {
    if (root==NULL) {
      *current=root= new Array(maxarraysize);
    } 
    for (i=0; i<keyLen; i++) {
      if ((elemptr=(*current)->Insert(keyList[i])) == NULL) {
	Debug(stderr, "Converting array to btree at level %d\n", i+1);
	fflush(stderr);
	tmp=new Btree(btreeary, (Array*)(*current));
	elemptr=tmp->Insert(keyList[i]);
	*current=tmp;
      }

      if ((*current)->getId()==btree && ((Btree*)(*current))->getHeight()>btreemaxheight) {
	tmp=new HashTable(numbucket, bucketsize);
	((Btree*)(*current))->toHtable((HashTable*)tmp);
	*current=tmp;
	elemptr=tmp->Search(keyList[i]);
      }

      if (i < keyLen -1 && elemptr->next==NULL) {
	elemptr->next=new Array(maxarraysize);
      }
        
      current=&(elemptr->next);
    }
  }

  if (elemptr->data !=-1)  //data already there
    return False;
  else {
    elemptr->data=d;
    return True;
  }
  
}

Boolean Trie::Delete(int keyLen, keyType *keyList) {
  return False;
}

dataType *Trie::Search(int keyLen, keyType *keyList) {
  TrieNode* current=root; INS_INC(2);
  elementType* elemptr; 
  for (register int i=0; i<keyLen; i++) { INS_INC(1); INS_INC(4); 
    INS_INC(3);
    if (current==NULL) return NULL; INS_INC(2);
    elemptr=current->Search(keyList[i]); INS_INC(16);
    INS_INC(3);
    if (elemptr==NULL) return NULL; INS_INC(2);
    current=elemptr->next; INS_INC(3);
  }
  INS_INC(2);
  if (elemptr->data==-1) return NULL;
  else return &(elemptr->data); 
}

void Trie::print_klist(FILE *dumpFp) {
  for (int i=0; i<=depth; i++) 
    fprintf(dumpFp, " %d", elemptr[i]->key);
  fprintf(dumpFp, " --> %d\n", elemptr[depth]->data);
}

void Trie::Dump(FILE *dumpFp) {
  tnList[0]=root;
  elemptr[0]=root->getFirst();
  depth=0;

  while (1) {
    if (elemptr[depth]->data!=-1) print_klist(dumpFp);

    if (elemptr[depth]->next!=NULL) {
      tnList[depth+1]=elemptr[depth]->next;
      depth++;
      elemptr[depth]=tnList[depth]->getFirst();
    }
    else {
      while (depth>=0 && ((elemptr[depth]=tnList[depth]->getNext())==NULL))
	depth--;
      if (depth<0) break;
    }
  }
}
